#include <stdlib.h>
#include <stdio.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/pgmspace.h>

#include "vari_code.h"			//lookup table

#define PACKETSIZE 90

#define SET_DTR sbi(PORTD,1)
#define CLEAR_DTR cbi(PORTD,1)

#define RADIO_ON sbi(PORTB,2)
#define RADIO_OFF cbi(PORTB,2)
									//	\/ baud rate
#define BAUDDIV  (u16)( ((float)F_CPU/(9600*16UL)) -1 )//baud divider

#define BV(bit)			(1<<(bit))
#define cbi(reg,bit)	reg &= ~(BV(bit))
#define sbi(reg,bit)	reg |= (BV(bit))
#define	outb(addr, data)	addr = (data)
#define	inb(addr)			(addr)

typedef unsigned char u08;
typedef signed char s08;
typedef unsigned int u16;
typedef unsigned long u32;
typedef struct { u08 bytes[3]; u08 size;} varicoder_type;
extern const unsigned short varicode_lookup[256];//the lookup table for varicode - the 4 msb are the number of valid bits in the data
volatile u08 bufferfront,bufferback,buffer[256],PWM_busy_flag=0,PWM;
void set_PWM(u08 set);
void put_char(char c, FILE * stream);
void get_char(FILE * stream);
FILE uart_stdout = FDEV_SETUP_STREAM(put_char,get_char, _FDEV_SETUP_RW);

u08 gray_encode(u08 arg)		//gray encodes a nibble
{
	u08 g=arg&0x08,b;			//takes a hex argument - i.e. 4 bit Gray coder
	for (g|=0x30;g&0xF0;g-=0x10)//we reuse the upper 4 msb of g as a counter
	{
		b = ((arg&(1<<(g>>4)))>>1) ^ (arg&(1<<((g>>4)-1)));
		g|=b;
	}
	return g&0x0F;
}

u16 FEC(u08 input)				//r=1/2, k=7 convolution encodes a u08 to give a u16
{
	static u16 reg;
	u08 n;
	u32 output=0;				//lsb is defined as earliest in time
	reg|=((u16)input)<<8;		//load the input into the top of the register
	for(n=16;n<32;n++)
	{
		output|=((u32)(reg&(1<<0))<<n)^((u32)(reg&(1<<2))<<(n-2))^((u32)(reg&(1<<3))<<(n-3))^((u32)(reg&(1<<5))<<(n-5))^((u32)(reg&(1<<6))<<(n-6));//first tap output
		n++;
		output|=((u32)(reg&(1<<0))<<n)^((u32)(reg&(1<<1))<<(n-1))^((u32)(reg&(1<<2))<<(n-2))^((u32)(reg&(1<<3))<<(n-3))^((u32)(reg&(1<<6))<<(n-6));//second tap output	
		reg>>=1;				//shift the register along
	}
	return output>>16;			//the two streams, interleaved
}

varicoder_type varicode(u08 ascii_in)//impliments a varicode buffer
{
	static s08 top;
	static u32 internal_buffer;
	varicoder_type output;	
	u16 bits=pgm_read_word(&varicode_lookup[ascii_in]);//the upper 4 msb tell us how many valid bits there are for this character
	internal_buffer|=(bits&0x0FFF)<<top;
	top+=(bits&0xF000)>>12;		//number of bits we collected in varicode
	for(output.size=0;top>7;top-=8,internal_buffer>>=8,output.size++)//loop to get bytes out of the buffer
		output.bytes[output.size]=(u08)internal_buffer;
	return output;
}

u08 interleaver(u08 arg)		//interleaves nibble arguments with internal buffer
{
	static u08 buffer[20],n;	//n is our position in the buffer in bits
	u16 l;
	for(l=0;l<4;l++)			//loop through sticking a bit in each 41 bit section of the 160 bit buffer
	{
		buffer[n>>3]&=~(1<<(n&0x07));//set the bit to be swapped to zero
        buffer[n>>3]|=(arg&1)<<(n&0x07);//set a bit in each of the 4 41 bit sections to the input bit
		arg>>=1;				//get the next bit of arg
		n+=41;
		n%=160;
	}
	l=(buffer[n>>3] | (buffer[(((n+4)%160)>>3)]<<8) )>>(n&0x07);//create a u16 from two consecutive elements of the buffer, then shift down
	return (u08)(0x000F&l);		//use the lower 4 lsb as a nibble
}
	
void process_ascii(u08 ascii)	//takes a ascii character and modulates the radio using the output stream - blocks while data is sent
{
	varicoder_type v=varicode(ascii);//varicode our ascii char
	u16 encoded;
	u08 n,m;
	for(n=0;v.size;v.size--,n++)//for each byte returned from the varicoder
	{
		encoded=FEC(v.bytes[n]);//loop through the returned bytes and put them into the FEC
		for(m=0;m<4;m++)		//each FEC encoded byte is 4 nibbles
		{
			PWM=gray_encode(interleaver(encoded&0x000F));//set the global PWM variable
			while(PWM_busy_flag);//wait for the PWM to be free
			PWM_busy_flag=1;	//pwm is now busy
			encoded>>=4;		//get the next nibble
			putchar(0x30+PWM);	//what we sent
			putchar(0x2C);		//comma
		}
	}
}

void put_char(char c, FILE * stream)
{
	while ( !( UCSR0A & (1<<UDRE0)) );
	UDR0=c;
}

void get_char(FILE * stream)
{;}

ISR(USART_RX_vect)
{
	buffer[bufferfront]=UDR0;	//stick the incoming data in the buffer
	bufferfront++;
	if(bufferfront==bufferback)
		bufferback++;			//our buffer was filled
	if((u08)((s08)bufferback-(s08)bufferfront)<PACKETSIZE)//flow control - clear if not enough space
		CLEAR_DTR;
}

ISR(TIMER1_COMPA_vect)			//fires at the same time as end of pwm pulse
{
	static u16 counter;
	counter++;
	if(counter==750)			//1/62.5 seconds at 12MHz
	{
		counter=0;
		PWM_busy_flag=0;
		OCR1AL=(PWM<<4);		//set the pwm
	}
}

int main(void)
{
	power_adc_disable();		//turn off unneeded modules
	power_spi_disable();
	power_timer0_disable();
	power_timer2_disable();
	power_twi_disable();
	DDRB=0b00000110;			//radio enable and pwm pins
	DDRD=0b00000010;			//use TX as a DTR pin
	outb(UCSR0B, BV(RXCIE0)|BV(RXEN0)|BV(TXEN0));//mega xx8 registers - enable rx with interrupts
	outb(UCSR0C, BV(UPM01)|BV(UPM00)|BV(UCSZ01)|BV(UCSZ00));//setup 8O1
	outb(UBRR0L, BAUDDIV);
	outb(UBRR0H, BAUDDIV>>8);	//end of UART setup
	//Timer 1 setup
	sbi(TCCR1B,CS10);			//clock at F_CPU
	TCNT1 = 0;					//reset timer1  
 	// set fast PWM mode with ICR top-count
	cbi(TCCR1A,WGM10);
	sbi(TCCR1A,WGM11);
	sbi(TCCR1B,WGM12);
	sbi(TCCR1B,WGM13);
 	// set top count value to give correct period PWM - 256
	ICR1 = 255;
 	// clear output compare value A
	OCR1A = 0;
	// turn on channel A (OC1A) PWM output
	// set OC1A as non-inverted PWM
	sbi(TCCR1A,COM1A1);
	cbi(TCCR1A,COM1A0);
	//enable compare 1A interrupts
	outb(TIMSK1, BV(OCIE1A));	//end of pwm setup
	wdt_enable(WDTO_4S);
	stdout=&uart_stdout;
	//enable the radio
	SET_DTR;
	RADIO_ON;
	__asm__ __volatile__ ("sei" ::);//enable global interrupts
	for(;;)						//main loop
	{
		if(bufferback==bufferfront)
			process_ascii(0);	//send null to keep the rx DLL ect in sinc
		else
		{
			process_ascii(buffer[bufferback]);//send something from the buffer
			bufferback++;		//move to next buffer location		
		}
		wdt_reset();
		if((u08)((s08)bufferback-(s08)bufferfront)>PACKETSIZE)//flow control - set if we have space
			SET_DTR;
	}
}
